@@ -61,10 +61,9 @@ PyObject *normalizestring(const char *string)
         return NULL;
     }
 
-    v = PyString_FromStringAndSize(NULL, len);
-    if (v == NULL)
+    p = PyMem_Malloc(len + 1);
+    if (p == NULL)
         return NULL;
-    p = PyString_AS_STRING(v);
     for (i = 0; i < len; i++) {
         register char ch = string[i];
         if (ch == ' ')
@@ -73,6 +72,11 @@ PyObject *normalizestring(const char *string)
             ch = Py_TOLOWER(Py_CHARMASK(ch));
         p[i] = ch;
     }
+    p[i] = '\0';
+    v = PyUnicode_FromString(p);
+    if (v == NULL)
+        return NULL;
+    PyMem_Free(p);
     return v;
 }
 
@@ -112,7 +116,7 @@ PyObject *_PyCodec_Lookup(const char *encoding)
     v = normalizestring(encoding);
     if (v == NULL)
         goto onError;
-    PyString_InternInPlace(&v);
+    PyUnicode_InternInPlace(&v);
 
     /* First, try to lookup the name in the registry dictionary */
     result = PyDict_GetItem(interp->codec_search_cache, v);
@@ -167,7 +171,10 @@ PyObject *_PyCodec_Lookup(const char *encoding)
     }
 
     /* Cache and return the result */
-    PyDict_SetItem(interp->codec_search_cache, v, result);
+    if (PyDict_SetItem(interp->codec_search_cache, v, result) < 0) {
+        Py_DECREF(result);
+        goto onError;
+    }
     Py_DECREF(args);
     return result;
 
@@ -176,6 +183,23 @@ PyObject *_PyCodec_Lookup(const char *encoding)
     return NULL;
 }
 
+/* Codec registry encoding check API. */
+
+int PyCodec_KnownEncoding(const char *encoding)
+{
+    PyObject *codecs;
+
+    codecs = _PyCodec_Lookup(encoding);
+    if (!codecs) {
+        PyErr_Clear();
+        return 0;
+    }
+    else {
+        Py_DECREF(codecs);
+        return 1;
+    }
+}
+
 static
 PyObject *args_tuple(PyObject *object,
                      const char *errors)
@@ -190,7 +214,7 @@ PyObject *args_tuple(PyObject *object,
     if (errors) {
         PyObject *v;
 
-        v = PyString_FromString(errors);
+        v = PyUnicode_FromString(errors);
         if (v == NULL) {
             Py_DECREF(args);
             return NULL;
@@ -217,15 +241,20 @@ PyObject *codec_getitem(const char *encoding, int index)
     return v;
 }
 
-/* Helper functions to create an incremental codec. */
-static
-PyObject *codec_makeincrementalcodec(PyObject *codec_info,
-                                     const char *errors,
-                                     const char *attrname)
-{
-    PyObject *ret, *inccodec;
+/* Helper function to create an incremental codec. */
 
-    inccodec = PyObject_GetAttrString(codec_info, attrname);
+static
+PyObject *codec_getincrementalcodec(const char *encoding,
+                                    const char *errors,
+                                    const char *attrname)
+{
+    PyObject *codecs, *ret, *inccodec;
+
+    codecs = _PyCodec_Lookup(encoding);
+    if (codecs == NULL)
+        return NULL;
+    inccodec = PyObject_GetAttrString(codecs, attrname);
+    Py_DECREF(codecs);
     if (inccodec == NULL)
         return NULL;
     if (errors)
@@ -236,21 +265,6 @@ PyObject *codec_makeincrementalcodec(PyObject *codec_info,
     return ret;
 }
 
-static
-PyObject *codec_getincrementalcodec(const char *encoding,
-                                    const char *errors,
-                                    const char *attrname)
-{
-    PyObject *codec_info, *ret;
-
-    codec_info = _PyCodec_Lookup(encoding);
-    if (codec_info == NULL)
-        return NULL;
-    ret = codec_makeincrementalcodec(codec_info, errors, attrname);
-    Py_DECREF(codec_info);
-    return ret;
-}
-
 /* Helper function to create a stream codec. */
 
 static
@@ -274,24 +288,6 @@ PyObject *codec_getstreamcodec(const char *encoding,
     return streamcodec;
 }
 
-/* Helpers to work with the result of _PyCodec_Lookup
-
- */
-PyObject *_PyCodecInfo_GetIncrementalDecoder(PyObject *codec_info,
-                                             const char *errors)
-{
-    return codec_makeincrementalcodec(codec_info, errors,
-                                      "incrementaldecoder");
-}
-
-PyObject *_PyCodecInfo_GetIncrementalEncoder(PyObject *codec_info,
-                                             const char *errors)
-{
-    return codec_makeincrementalcodec(codec_info, errors,
-                                      "incrementalencoder");
-}
-
-
 /* Convenience APIs to query the Codec registry.
 
    All APIs return a codec object with incremented refcount.
@@ -339,27 +335,30 @@ PyObject *PyCodec_StreamWriter(const char *encoding,
 
    errors is passed to the encoder factory as argument if non-NULL. */
 
-static PyObject *
-_PyCodec_EncodeInternal(PyObject *object,
-                        PyObject *encoder,
-                        const char *encoding,
-                        const char *errors)
+PyObject *PyCodec_Encode(PyObject *object,
+                         const char *encoding,
+                         const char *errors)
 {
+    PyObject *encoder = NULL;
     PyObject *args = NULL, *result = NULL;
-    PyObject *v;
+    PyObject *v = NULL;
+
+    encoder = PyCodec_Encoder(encoding);
+    if (encoder == NULL)
+        goto onError;
 
     args = args_tuple(object, errors);
     if (args == NULL)
         goto onError;
 
-    result = PyEval_CallObject(encoder,args);
+    result = PyEval_CallObject(encoder, args);
     if (result == NULL)
         goto onError;
 
     if (!PyTuple_Check(result) ||
         PyTuple_GET_SIZE(result) != 2) {
         PyErr_SetString(PyExc_TypeError,
-                        "encoder must return a tuple (object,integer)");
+                        "encoder must return a tuple (object, integer)");
         goto onError;
     }
     v = PyTuple_GET_ITEM(result,0);
@@ -383,15 +382,18 @@ _PyCodec_EncodeInternal(PyObject *object,
 
    errors is passed to the decoder factory as argument if non-NULL. */
 
-static PyObject *
-_PyCodec_DecodeInternal(PyObject *object,
-                        PyObject *decoder,
-                        const char *encoding,
-                        const char *errors)
+PyObject *PyCodec_Decode(PyObject *object,
+                         const char *encoding,
+                         const char *errors)
 {
+    PyObject *decoder = NULL;
     PyObject *args = NULL, *result = NULL;
     PyObject *v;
 
+    decoder = PyCodec_Decoder(encoding);
+    if (decoder == NULL)
+        goto onError;
+
     args = args_tuple(object, errors);
     if (args == NULL)
         goto onError;
@@ -421,139 +423,6 @@ _PyCodec_DecodeInternal(PyObject *object,
     return NULL;
 }
 
-/* Generic encoding/decoding API */
-PyObject *PyCodec_Encode(PyObject *object,
-                         const char *encoding,
-                         const char *errors)
-{
-    PyObject *encoder;
-
-    encoder = PyCodec_Encoder(encoding);
-    if (encoder == NULL)
-        return NULL;
-
-    return _PyCodec_EncodeInternal(object, encoder, encoding, errors);
-}
-
-PyObject *PyCodec_Decode(PyObject *object,
-                         const char *encoding,
-                         const char *errors)
-{
-    PyObject *decoder;
-
-    decoder = PyCodec_Decoder(encoding);
-    if (decoder == NULL)
-        return NULL;
-
-    return _PyCodec_DecodeInternal(object, decoder, encoding, errors);
-}
-
-/* Text encoding/decoding API */
-PyObject * _PyCodec_LookupTextEncoding(const char *encoding,
-                                       const char *alternate_command)
-{
-    PyObject *codec;
-    PyObject *attr;
-    int is_text_codec;
-
-    codec = _PyCodec_Lookup(encoding);
-    if (codec == NULL)
-        return NULL;
-
-    /* Backwards compatibility: assume any raw tuple describes a text
-     * encoding, and the same for anything lacking the private
-     * attribute.
-     */
-    if (Py_Py3kWarningFlag && !PyTuple_CheckExact(codec)) {
-        attr = PyObject_GetAttrString(codec, "_is_text_encoding");
-        if (attr == NULL) {
-            if (!PyErr_ExceptionMatches(PyExc_AttributeError))
-                goto onError;
-            PyErr_Clear();
-        } else {
-            is_text_codec = PyObject_IsTrue(attr);
-            Py_DECREF(attr);
-            if (is_text_codec < 0)
-                goto onError;
-            if (!is_text_codec) {
-                PyObject *msg = PyString_FromFormat(
-                            "'%.400s' is not a text encoding; "
-                            "use %s to handle arbitrary codecs",
-                            encoding, alternate_command);
-                if (msg == NULL)
-                    goto onError;
-                if (PyErr_WarnPy3k(PyString_AS_STRING(msg), 1) < 0) {
-                    Py_DECREF(msg);
-                    goto onError;
-                }
-                Py_DECREF(msg);
-            }
-        }
-    }
-
-    /* This appears to be a valid text encoding */
-    return codec;
-
- onError:
-    Py_DECREF(codec);
-    return NULL;
-}
-
-
-static
-PyObject *codec_getitem_checked(const char *encoding,
-                                const char *alternate_command,
-                                int index)
-{
-    PyObject *codec;
-    PyObject *v;
-
-    codec = _PyCodec_LookupTextEncoding(encoding, alternate_command);
-    if (codec == NULL)
-        return NULL;
-
-    v = PyTuple_GET_ITEM(codec, index);
-    Py_INCREF(v);
-    Py_DECREF(codec);
-    return v;
-}
-
-static PyObject * _PyCodec_TextEncoder(const char *encoding)
-{
-    return codec_getitem_checked(encoding, "codecs.encode()", 0);
-}
-
-static PyObject * _PyCodec_TextDecoder(const char *encoding)
-{
-    return codec_getitem_checked(encoding, "codecs.decode()", 1);
-}
-
-PyObject *_PyCodec_EncodeText(PyObject *object,
-                              const char *encoding,
-                              const char *errors)
-{
-    PyObject *encoder;
-
-    encoder = _PyCodec_TextEncoder(encoding);
-    if (encoder == NULL)
-        return NULL;
-
-    return _PyCodec_EncodeInternal(object, encoder, encoding, errors);
-}
-
-PyObject *_PyCodec_DecodeText(PyObject *object,
-                              const char *encoding,
-                              const char *errors)
-{
-    PyObject *decoder;
-
-    decoder = _PyCodec_TextDecoder(encoding);
-    if (decoder == NULL)
-        return NULL;
-
-    return _PyCodec_DecodeInternal(object, decoder, encoding, errors);
-}
-
 /* Register the error handling callback function error under the name
    name. This function will be called by the codec when it encounters
    an unencodable characters/undecodable bytes and doesn't know the
@@ -601,14 +470,9 @@ static void wrong_exception_type(PyObject *exc)
         PyObject *name = PyObject_GetAttrString(type, "__name__");
         Py_DECREF(type);
         if (name != NULL) {
-            PyObject *string = PyObject_Str(name);
+            PyErr_Format(PyExc_TypeError,
+                         "don't know how to handle %S in error callback", name);
             Py_DECREF(name);
-            if (string != NULL) {
-                PyErr_Format(PyExc_TypeError,
-                    "don't know how to handle %.400s in error callback",
-                    PyString_AS_STRING(string));
-                Py_DECREF(string);
-            }
         }
     }
 }
@@ -623,20 +487,18 @@ PyObject *PyCodec_StrictErrors(PyObject *exc)
 }
 
 
-#ifdef Py_USING_UNICODE
 PyObject *PyCodec_IgnoreErrors(PyObject *exc)
 {
     Py_ssize_t end;
-
-    if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeEncodeError)) {
+    if (PyObject_IsInstance(exc, PyExc_UnicodeEncodeError)) {
         if (PyUnicodeEncodeError_GetEnd(exc, &end))
             return NULL;
     }
-    else if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeDecodeError)) {
+    else if (PyObject_IsInstance(exc, PyExc_UnicodeDecodeError)) {
         if (PyUnicodeDecodeError_GetEnd(exc, &end))
             return NULL;
     }
-    else if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeTranslateError)) {
+    else if (PyObject_IsInstance(exc, PyExc_UnicodeTranslateError)) {
         if (PyUnicodeTranslateError_GetEnd(exc, &end))
             return NULL;
     }
@@ -656,7 +518,7 @@ PyObject *PyCodec_ReplaceErrors(PyObject *exc)
     Py_ssize_t end;
     Py_ssize_t i;
 
-    if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeEncodeError)) {
+    if (PyObject_IsInstance(exc, PyExc_UnicodeEncodeError)) {
         PyObject *res;
         Py_UNICODE *p;
         if (PyUnicodeEncodeError_GetStart(exc, &start))
@@ -673,13 +535,13 @@ PyObject *PyCodec_ReplaceErrors(PyObject *exc)
         Py_DECREF(res);
         return restuple;
     }
-    else if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeDecodeError)) {
+    else if (PyObject_IsInstance(exc, PyExc_UnicodeDecodeError)) {
         Py_UNICODE res = Py_UNICODE_REPLACEMENT_CHARACTER;
         if (PyUnicodeDecodeError_GetEnd(exc, &end))
             return NULL;
-        return Py_BuildValue("(u#n)", &res, (Py_ssize_t)1, end);
+        return Py_BuildValue("(u#n)", &res, 1, end);
     }
-    else if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeTranslateError)) {
+    else if (PyObject_IsInstance(exc, PyExc_UnicodeTranslateError)) {
         PyObject *res;
         Py_UNICODE *p;
         if (PyUnicodeTranslateError_GetStart(exc, &start))
@@ -704,7 +566,7 @@ PyObject *PyCodec_ReplaceErrors(PyObject *exc)
 
 PyObject *PyCodec_XMLCharRefReplaceErrors(PyObject *exc)
 {
-    if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeEncodeError)) {
+    if (PyObject_IsInstance(exc, PyExc_UnicodeEncodeError)) {
         PyObject *restuple;
         PyObject *object;
         Py_ssize_t start;
@@ -712,9 +574,8 @@ PyObject *PyCodec_XMLCharRefReplaceErrors(PyObject *exc)
         PyObject *res;
         Py_UNICODE *p;
         Py_UNICODE *startp;
-        Py_UNICODE *e;
         Py_UNICODE *outp;
-        Py_ssize_t ressize;
+        int ressize;
         if (PyUnicodeEncodeError_GetStart(exc, &start))
             return NULL;
         if (PyUnicodeEncodeError_GetEnd(exc, &end))
@@ -722,38 +583,26 @@ PyObject *PyCodec_XMLCharRefReplaceErrors(PyObject *exc)
         if (!(object = PyUnicodeEncodeError_GetObject(exc)))
             return NULL;
         startp = PyUnicode_AS_UNICODE(object);
-        if (end - start > PY_SSIZE_T_MAX / (2+7+1)) {
-            end = start + PY_SSIZE_T_MAX / (2+7+1);
-#ifndef Py_UNICODE_WIDE
-            if (0xD800 <= startp[end - 1] && startp[end - 1] <= 0xDBFF)
-                end--;
-#endif
-        }
-        e = startp + end;
-        for (p = startp+start, ressize = 0; p < e;) {
-            Py_UCS4 ch = *p++;
-#ifndef Py_UNICODE_WIDE
-            if ((0xD800 <= ch && ch <= 0xDBFF) &&
-                (p < e) &&
-                (0xDC00 <= *p && *p <= 0xDFFF)) {
-                ch = ((((ch & 0x03FF) << 10) |
-                       ((Py_UCS4)*p++ & 0x03FF)) + 0x10000);
-            }
-#endif
-            if (ch < 10)
+        for (p = startp+start, ressize = 0; p < startp+end; ++p) {
+            if (*p<10)
                 ressize += 2+1+1;
-            else if (ch < 100)
+            else if (*p<100)
                 ressize += 2+2+1;
-            else if (ch < 1000)
+            else if (*p<1000)
                 ressize += 2+3+1;
-            else if (ch < 10000)
+            else if (*p<10000)
                 ressize += 2+4+1;
-            else if (ch < 100000)
+#ifndef Py_UNICODE_WIDE
+            else
                 ressize += 2+5+1;
-            else if (ch < 1000000)
+#else
+            else if (*p<100000)
+                ressize += 2+5+1;
+            else if (*p<1000000)
                 ressize += 2+6+1;
             else
                 ressize += 2+7+1;
+#endif
         }
         /* allocate replacement */
         res = PyUnicode_FromUnicode(NULL, ressize);
@@ -762,41 +611,40 @@ PyObject *PyCodec_XMLCharRefReplaceErrors(PyObject *exc)
             return NULL;
         }
         /* generate replacement */
-        for (p = startp+start, outp = PyUnicode_AS_UNICODE(res); p < e;) {
+        for (p = startp+start, outp = PyUnicode_AS_UNICODE(res);
+            p < startp+end; ++p) {
+            Py_UNICODE c = *p;
             int digits;
             int base;
-            Py_UCS4 ch = *p++;
-#ifndef Py_UNICODE_WIDE
-            if ((0xD800 <= ch && ch <= 0xDBFF) &&
-                (p < startp+end) &&
-                (0xDC00 <= *p && *p <= 0xDFFF)) {
-                ch = ((((ch & 0x03FF) << 10) |
-                       ((Py_UCS4)*p++ & 0x03FF)) + 0x10000);
-            }
-#endif
             *outp++ = '&';
             *outp++ = '#';
-            if (ch < 10) {
+            if (*p<10) {
                 digits = 1;
                 base = 1;
             }
-            else if (ch < 100) {
+            else if (*p<100) {
                 digits = 2;
                 base = 10;
             }
-            else if (ch < 1000) {
+            else if (*p<1000) {
                 digits = 3;
                 base = 100;
             }
-            else if (ch < 10000) {
+            else if (*p<10000) {
                 digits = 4;
                 base = 1000;
             }
-            else if (ch < 100000) {
+#ifndef Py_UNICODE_WIDE
+            else {
                 digits = 5;
                 base = 10000;
             }
-            else if (ch < 1000000) {
+#else
+            else if (*p<100000) {
+                digits = 5;
+                base = 10000;
+            }
+            else if (*p<1000000) {
                 digits = 6;
                 base = 100000;
             }
@@ -804,9 +652,10 @@ PyObject *PyCodec_XMLCharRefReplaceErrors(PyObject *exc)
                 digits = 7;
                 base = 1000000;
             }
+#endif
             while (digits-->0) {
-                *outp++ = '0' + ch/base;
-                ch %= base;
+                *outp++ = '0' + c/base;
+                c %= base;
                 base /= 10;
             }
             *outp++ = ';';
@@ -829,7 +678,14 @@ static Py_UNICODE hexdigits[] = {
 
 PyObject *PyCodec_BackslashReplaceErrors(PyObject *exc)
 {
-    if (PyObject_TypeCheck(exc, (PyTypeObject *)PyExc_UnicodeEncodeError)) {
+#ifndef Py_UNICODE_WIDE
+#define IS_SURROGATE_PAIR(p, end) \
+    (*p >= 0xD800 && *p <= 0xDBFF && (p + 1) < end && \
+     *(p + 1) >= 0xDC00 && *(p + 1) <= 0xDFFF)
+#else
+#define IS_SURROGATE_PAIR(p, end) 0
+#endif
+    if (PyObject_IsInstance(exc, PyExc_UnicodeEncodeError)) {
         PyObject *restuple;
         PyObject *object;
         Py_ssize_t start;
@@ -838,15 +694,13 @@ PyObject *PyCodec_BackslashReplaceErrors(PyObject *exc)
         Py_UNICODE *p;
         Py_UNICODE *startp;
         Py_UNICODE *outp;
-        Py_ssize_t ressize;
+        int ressize;
         if (PyUnicodeEncodeError_GetStart(exc, &start))
             return NULL;
         if (PyUnicodeEncodeError_GetEnd(exc, &end))
             return NULL;
         if (!(object = PyUnicodeEncodeError_GetObject(exc)))
             return NULL;
-        if (end - start > PY_SSIZE_T_MAX / (1+1+8))
-            end = start + PY_SSIZE_T_MAX / (1+1+8);
         startp = PyUnicode_AS_UNICODE(object);
         for (p = startp+start, ressize = 0; p < startp+end; ++p) {
 #ifdef Py_UNICODE_WIDE
@@ -855,21 +709,27 @@ PyObject *PyCodec_BackslashReplaceErrors(PyObject *exc)
             else
 #endif
             if (*p >= 0x100) {
-                ressize += 1+1+4;
+                if (IS_SURROGATE_PAIR(p, startp+end)) {
+                    ressize += 1+1+8;
+                    ++p;
+                }
+                else
+                    ressize += 1+1+4;
             }
             else
                 ressize += 1+1+2;
         }
         res = PyUnicode_FromUnicode(NULL, ressize);
-        if (res == NULL) {
-            Py_DECREF(object);
+        if (res==NULL)
             return NULL;
-        }
         for (p = startp+start, outp = PyUnicode_AS_UNICODE(res);
             p < startp+end; ++p) {
-            Py_UNICODE c = *p;
+            Py_UCS4 c = (Py_UCS4) *p;
             *outp++ = '\\';
-#ifdef Py_UNICODE_WIDE
+            if (IS_SURROGATE_PAIR(p, startp+end)) {
+                c = ((*p & 0x3FF) << 10) + (*(p + 1) & 0x3FF) + 0x10000;
+                ++p;
+            }
             if (c >= 0x00010000) {
                 *outp++ = 'U';
                 *outp++ = hexdigits[(c>>28)&0xf];
@@ -879,9 +739,7 @@ PyObject *PyCodec_BackslashReplaceErrors(PyObject *exc)
                 *outp++ = hexdigits[(c>>12)&0xf];
                 *outp++ = hexdigits[(c>>8)&0xf];
             }
-            else
-#endif
-            if (c >= 0x100) {
+            else if (c >= 0x100) {
                 *outp++ = 'u';
                 *outp++ = hexdigits[(c>>12)&0xf];
                 *outp++ = hexdigits[(c>>8)&0xf];
@@ -901,8 +759,166 @@ PyObject *PyCodec_BackslashReplaceErrors(PyObject *exc)
         wrong_exception_type(exc);
         return NULL;
     }
+#undef IS_SURROGATE_PAIR
 }
-#endif
+
+/* This handler is declared static until someone demonstrates
+   a need to call it directly. */
+static PyObject *
+PyCodec_SurrogatePassErrors(PyObject *exc)
+{
+    PyObject *restuple;
+    PyObject *object;
+    Py_ssize_t start;
+    Py_ssize_t end;
+    PyObject *res;
+    if (PyObject_IsInstance(exc, PyExc_UnicodeEncodeError)) {
+        Py_UNICODE *p;
+        Py_UNICODE *startp;
+        char *outp;
+        if (PyUnicodeEncodeError_GetStart(exc, &start))
+            return NULL;
+        if (PyUnicodeEncodeError_GetEnd(exc, &end))
+            return NULL;
+        if (!(object = PyUnicodeEncodeError_GetObject(exc)))
+            return NULL;
+        startp = PyUnicode_AS_UNICODE(object);
+        res = PyBytes_FromStringAndSize(NULL, 3*(end-start));
+        if (!res) {
+            Py_DECREF(object);
+            return NULL;
+        }
+        outp = PyBytes_AsString(res);
+        for (p = startp+start; p < startp+end; p++) {
+            Py_UNICODE ch = *p;
+            if (ch < 0xd800 || ch > 0xdfff) {
+                /* Not a surrogate, fail with original exception */
+                PyErr_SetObject(PyExceptionInstance_Class(exc), exc);
+                Py_DECREF(res);
+                Py_DECREF(object);
+                return NULL;
+            }
+            *outp++ = (char)(0xe0 | (ch >> 12));
+            *outp++ = (char)(0x80 | ((ch >> 6) & 0x3f));
+            *outp++ = (char)(0x80 | (ch & 0x3f));
+        }
+        restuple = Py_BuildValue("(On)", res, end);
+        Py_DECREF(res);
+        Py_DECREF(object);
+        return restuple;
+    }
+    else if (PyObject_IsInstance(exc, PyExc_UnicodeDecodeError)) {
+        unsigned char *p;
+        Py_UNICODE ch = 0;
+        if (PyUnicodeDecodeError_GetStart(exc, &start))
+            return NULL;
+        if (!(object = PyUnicodeDecodeError_GetObject(exc)))
+            return NULL;
+        if (!(p = (unsigned char*)PyBytes_AsString(object))) {
+            Py_DECREF(object);
+            return NULL;
+        }
+        /* Try decoding a single surrogate character. If
+           there are more, let the codec call us again. */
+        p += start;
+        if ((p[0] & 0xf0) == 0xe0 ||
+            (p[1] & 0xc0) == 0x80 ||
+            (p[2] & 0xc0) == 0x80) {
+            /* it's a three-byte code */
+            ch = ((p[0] & 0x0f) << 12) + ((p[1] & 0x3f) << 6) + (p[2] & 0x3f);
+            if (ch < 0xd800 || ch > 0xdfff)
+                /* it's not a surrogate - fail */
+                ch = 0;
+        }
+        Py_DECREF(object);
+        if (ch == 0) {
+            PyErr_SetObject(PyExceptionInstance_Class(exc), exc);
+            return NULL;
+        }
+        return Py_BuildValue("(u#n)", &ch, 1, start+3);
+    }
+    else {
+        wrong_exception_type(exc);
+        return NULL;
+    }
+}
+
+static PyObject *
+PyCodec_SurrogateEscapeErrors(PyObject *exc)
+{
+    PyObject *restuple;
+    PyObject *object;
+    Py_ssize_t start;
+    Py_ssize_t end;
+    PyObject *res;
+    if (PyObject_IsInstance(exc, PyExc_UnicodeEncodeError)) {
+        Py_UNICODE *p;
+        Py_UNICODE *startp;
+        char *outp;
+        if (PyUnicodeEncodeError_GetStart(exc, &start))
+            return NULL;
+        if (PyUnicodeEncodeError_GetEnd(exc, &end))
+            return NULL;
+        if (!(object = PyUnicodeEncodeError_GetObject(exc)))
+            return NULL;
+        startp = PyUnicode_AS_UNICODE(object);
+        res = PyBytes_FromStringAndSize(NULL, end-start);
+        if (!res) {
+            Py_DECREF(object);
+            return NULL;
+        }
+        outp = PyBytes_AsString(res);
+        for (p = startp+start; p < startp+end; p++) {
+            Py_UNICODE ch = *p;
+            if (ch < 0xdc80 || ch > 0xdcff) {
+                /* Not a UTF-8b surrogate, fail with original exception */
+                PyErr_SetObject(PyExceptionInstance_Class(exc), exc);
+                Py_DECREF(res);
+                Py_DECREF(object);
+                return NULL;
+            }
+            *outp++ = ch - 0xdc00;
+        }
+        restuple = Py_BuildValue("(On)", res, end);
+        Py_DECREF(res);
+        Py_DECREF(object);
+        return restuple;
+    }
+    else if (PyObject_IsInstance(exc, PyExc_UnicodeDecodeError)) {
+        unsigned char *p;
+        Py_UNICODE ch[4]; /* decode up to 4 bad bytes. */
+        int consumed = 0;
+        if (PyUnicodeDecodeError_GetStart(exc, &start))
+            return NULL;
+        if (PyUnicodeDecodeError_GetEnd(exc, &end))
+            return NULL;
+        if (!(object = PyUnicodeDecodeError_GetObject(exc)))
+            return NULL;
+        if (!(p = (unsigned char*)PyBytes_AsString(object))) {
+            Py_DECREF(object);
+            return NULL;
+        }
+        while (consumed < 4 && consumed < end-start) {
+            /* Refuse to escape ASCII bytes. */
+            if (p[start+consumed] < 128)
+                break;
+            ch[consumed] = 0xdc00 + p[start+consumed];
+            consumed++;
+        }
+        Py_DECREF(object);
+        if (!consumed) {
+            /* codec complained about ASCII byte. */
+            PyErr_SetObject(PyExceptionInstance_Class(exc), exc);
+            return NULL;
+        }
+        return Py_BuildValue("(u#n)", ch, consumed, start+consumed);
+    }
+    else {
+        wrong_exception_type(exc);
+        return NULL;
+    }
+}
+
 
 static PyObject *strict_errors(PyObject *self, PyObject *exc)
 {
@@ -910,7 +926,6 @@ static PyObject *strict_errors(PyObject *self, PyObject *exc)
 }
 
 
-#ifdef Py_USING_UNICODE
 static PyObject *ignore_errors(PyObject *self, PyObject *exc)
 {
     return PyCodec_IgnoreErrors(exc);
@@ -933,7 +948,16 @@ static PyObject *backslashreplace_errors(PyObject *self, PyObject *exc)
 {
     return PyCodec_BackslashReplaceErrors(exc);
 }
-#endif
+
+static PyObject *surrogatepass_errors(PyObject *self, PyObject *exc)
+{
+    return PyCodec_SurrogatePassErrors(exc);
+}
+
+static PyObject *surrogateescape_errors(PyObject *self, PyObject *exc)
+{
+    return PyCodec_SurrogateEscapeErrors(exc);
+}
 
 static int _PyCodecRegistry_Init(void)
 {
@@ -952,7 +976,6 @@ static int _PyCodecRegistry_Init(void)
                           "raises a UnicodeError on coding errors.")
             }
         },
-#ifdef Py_USING_UNICODE
         {
             "ignore",
             {
@@ -994,8 +1017,23 @@ static int _PyCodecRegistry_Init(void)
                           "which replaces an unencodable character with a "
                           "backslashed escape sequence.")
             }
+        },
+        {
+            "surrogatepass",
+            {
+                "surrogatepass",
+                surrogatepass_errors,
+                METH_O
+            }
+        },
+        {
+            "surrogateescape",
+            {
+                "surrogateescape",
+                surrogateescape_errors,
+                METH_O
+            }
         }
-#endif
     };
 
     PyInterpreterState *interp = PyThreadState_GET()->interp;
@@ -1027,19 +1065,11 @@ static int _PyCodecRegistry_Init(void)
         interp->codec_error_registry == NULL)
         Py_FatalError("can't initialize codec registry");
 
-    mod = PyImport_ImportModuleLevel("encodings", NULL, NULL, NULL, 0);
+    mod = PyImport_ImportModuleNoBlock("encodings");
     if (mod == NULL) {
-        if (PyErr_ExceptionMatches(PyExc_ImportError)) {
-            /* Ignore ImportErrors... this is done so that
-               distributions can disable the encodings package. Note
-               that other errors are not masked, e.g. SystemErrors
-               raised to inform the user of an error in the Python
-               configuration are still reported back to the user. */
-            PyErr_Clear();
-            return 0;
-        }
         return -1;
     }
     Py_DECREF(mod);
+    interp->codecs_initialized = 1;
     return 0;
 }
